#!/venv/bin/python
# -*- coding: utf-8 -*-
"""
# Este script está preparado para ejecutarse poco después del boot y notificar
por email su situación en la red (LAN/WAN).
"""
import datetime as dt
import logging
import os
import psutil
import re
import smtplib
import socket
import subprocess
import sys
import time
import yaml

VERBOSE = True
WAIT_INIT = 60
TIME_PAUSE = 5

BASEDIR = os.path.dirname(os.path.abspath(__file__))
PATH_SECRETS = os.path.join(BASEDIR, 'secrets.yaml')

LOG_LEVEL = 'INFO'  # 'DEBUG'
LOG_FILE = os.path.join(BASEDIR, 'cronip.log')
EXT_IP_FILE = os.path.join(BASEDIR, 'external_ip.txt')

# Configuration
with open(PATH_SECRETS) as _file:
    SECRETS = yaml.load(_file.read())
EMAIL_TARGET = SECRETS['email_target']
EMAIL_SERVER_USER = SECRETS['email_server']
EMAIL_SERVER_PWD = SECRETS['email_server_pw']
EMAIL_INITS = SECRETS.get('email_inits', False)
EMAIL_DEBUG = None
if 'debug_mode' in SECRETS:
    EMAIL_DEBUG = SECRETS['debug_mode']

def _get_cmd_output(cmd, default=None, verbose=True, **kwargs):
    list_cmd = cmd.split()
    # kwargs.update({'stdout': subprocess.PIPE})
    try:
        out = subprocess.check_output(list_cmd, **kwargs).decode()
        if out.endswith('\n'):
            return True, out[:-1]
        return True, out
    except subprocess.TimeoutExpired as e:
        time.sleep(TIME_PAUSE / 2)
        if verbose:
            print(
                '\nERROR subprocess.CalledProcessError: {} invocando el comando: {}'.format(
                    e, cmd))
    except subprocess.CalledProcessError as e:
        if verbose:
            print(
                '\nERROR subprocess.CalledProcessError: {} invocando el comando: {}'.format(
                    e, cmd))
    except FileNotFoundError as e:
        if verbose:
            print(
                '\nERROR FileNotFoundError: {} invocando el comando: {}'.format(
                    e, cmd))
    if default:
        return False, default
    return False, 0


def send_email(email_subject, email_content, target):
    try:
        smtpserver = smtplib.SMTP("smtp.gmail.com", 587)
        smtpserver.ehlo()
        smtpserver.starttls()
        smtpserver.login(EMAIL_SERVER_USER, EMAIL_SERVER_PWD)
        email = 'To: {}\nFrom: {}\nSubject: {}\n\n{}\n'.format(
            target, EMAIL_SERVER_USER, email_subject, email_content)
        smtpserver.sendmail(EMAIL_SERVER_USER, target, email.encode())
        smtpserver.close()
        logging.info('EMAIL SENDED={}'.format(email_subject))
        return True
    except smtplib.SMTPException as e:
        logging.error(
            'EMAIL NOT SENDED={}; SMTPException: {}'.format(email_subject, e))
    except socket.gaierror as e:
        logging.error(
            'EMAIL NOT SENDED={}; socket.gaierror: {}; '
            '¿Hay conexión a internet??'.format(email_subject, e))
    except Exception as e:
        logging.error(
            'EMAIL NOT SENDED={}; Exception: {}'.format(email_subject, e))
    return False


def _notify_email(d_result, extra_info=False, with_email=False):
    email_subject = 'RPI {} started! [IP: {}, {}]'.format(
        d_result['name'], d_result['ip_ext'], d_result['IP'])
    confs = ['name', 'user', 'path_home', 'IP']
    info_host = '\n'.join(
        ['  -> {:10} : {}'.format(k.upper(), d_result[k]) for k in confs])
    if extra_info:
        template = 'Configuración:\n{}\n\n**IP EXTERNA**: {}\n' \
                   '(Obtención en {:.3f} segs)\n\nIFCONFIG:\n{}'
        email_content = template.format(
            info_host, d_result['ip_ext'],
            d_result['time_ip_ext'], d_result['ifconfig'])
    else:
        template = 'Configuración:\n{}\n\n**IP EXTERNA**: {}\n' \
                   '(Obtención en {:.3f} segs)'
        email_content = template.format(
            info_host, d_result['ip_ext'], d_result['time_ip_ext'])
    if 'location_info' in d_result:
        email_content += '\n\nLocation:\t{}\n'.format(
            d_result['location_info'])

    ok_email = True
    if with_email and (EMAIL_TARGET is not None) and EMAIL_INITS:
        ok_email = send_email(email_subject, email_content, EMAIL_TARGET)
        if ok_email:
            logging.debug(
                'SALIDA OK de name:{}, file:{}'.format(__name__, __file__))
        else:
            logging.debug(
                'SALIDA SIN NOTIFICAR POR EMAIL de name:{}, file:{}'.format(
                    __name__, __file__))
    if ok_email and (EMAIL_DEBUG is not None):
        send_email(email_subject, email_content, EMAIL_DEBUG)


def get_machine_info(verbose):
    # Machine
    # NAME, IP, USER, IS_MAC, NUM_CPU, BOOT_TIME
    regexpr_ifconfig = re.compile(
        r'inet (addr:)?(?P<IP>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) ')
    # regexpr_IP = re.compile(r'(?P<IP1>\d{1,3})\.(?P<IP2>\d{1,3})\.
    # (?P<IP3>\d{1,3})\.(?P<IP4>\d{1,3})')

    def _extract_ifconfig(res_ifconfig):
        # IP(s) de la máquina local:
        busqueda = regexpr_ifconfig.findall(res_ifconfig)
        ips_locales = [b[1] for b in busqueda if b[1] != '127.0.0.1']
        return ips_locales

    # Nombre, tipo, user de la máquina local:
    is_mac = True if sys.platform == 'darwin' else False
    ok_name, name = _get_cmd_output('/bin/hostname', default='rpi')
    ok_current_user, current_user = _get_cmd_output(
        '/usr/bin/id -u -n', default='pi')
    # IP(s) de la máquina local:
    ok_ifconfig_output, ifconfig_output = _get_cmd_output(
        '/sbin/ifconfig', default='IFCONFIG No disponible')
    # ok_iwconfig_output, iwconfig_output = _get_cmd_output('/sbin/iwconfig wlan0',
    #  default='IWCONFIG No disponible', verbose=False)
    ip, hay_conexion, ips_lan = '127.0.0.1', False, []
    if ok_ifconfig_output:
        ips_lan = _extract_ifconfig(ifconfig_output)
        if len(ips_lan) > 0:
            ip = ips_lan[0]
            hay_conexion = True
    dict_machine = {'is_mac': is_mac, 'name': name, 'user': current_user,
                    'hay_conexion': hay_conexion, 'IP': ip,
                    'IPs': ips_lan, 'n_IPs': len(ips_lan),
                    'nCPU': (
                    psutil.cpu_count(), psutil.cpu_count(logical=False)),
                    'boot_time': dt.datetime.fromtimestamp(psutil.boot_time())}
    # if ok_iwconfig_output:
    #     dict_machine['iwconfig'] = iwconfig_output
    dict_machine['ifconfig'] = ifconfig_output
    # dict_machine['usb_devices'] = get_usb_devices()
    if verbose:
        [print('{0}:\t{1}'.format(k, v)) for k, v in dict_machine.items()]
    return dict_machine


def get_ext_ip(timeout=10):
    tic = time.time()
    n_retry = 0
    ok, ip, time_obtain, location_info = False, '', 1e6, None
    while n_retry < 2:
        # ok, ip_ext = _get_cmd_output('/usr/bin/curl ifconfig.me', default=None, timeout=timeout)
        # ok, ip_ext = _get_cmd_output('wget http://ipinfo.io/ip -qO -', default=None, timeout=timeout)

        ok = location_info is not None
        if ok:
            ip_ext = location_info.ip
        # print('Intento de ip_ext: {}'.format(ip_ext))
        toc = time.time()
        if ok:
            # ip = ip_ext[:-1]
            ip = ip_ext
        time_obtain = toc - tic

        try:
            with open(EXT_IP_FILE) as file_ip:
                old_ip = file_ip.read()
        except FileNotFoundError:
            old_ip = ''

        if ok and (not old_ip or old_ip != ip_ext):
            with open(EXT_IP_FILE, 'w') as file_ip:
                file_ip.write(ip_ext)
                mascara = 'WRITING NEW EXT_IP in {}: {}. T_OBTENCIÓN: {:.2f} seg'
        elif not ok:
            mascara = 'ERROR EXT_IP ({}) TIMEOUT?: {}. T_OBTENCIÓN: {:.2f} seg'
            n_retry += 1
        else:
            mascara = 'OK EXT_IP ({}) NOT CHANGING: {}. T_OBTENCIÓN: {:.2f} seg'
            n_retry = 1000
        logging.debug(mascara.format(EXT_IP_FILE, ip, time_obtain))
    return ok, ip, location_info, time_obtain


if __name__ == "__main__":
    import argparse

    p = argparse.ArgumentParser(description='RPI START NOTIFY SERVICE')
    p.add_argument('-s', '--silent', action='store_true', help='HIDES OUTPUT')
    p.add_argument('-e', '--email', action='store_true', help='SENDS EMAIL')
    p.add_argument('-p', '--print', action='store_true', help='PRINT LOG FILE')
    p.add_argument('--extra', action='store_true', help='REPORT EXTRA INFO')
    p.add_argument('--delete', action='store_true', help='DELETE LOG FILE')
    p.add_argument('--nowait', action='store_true',
                   help="DON'T WAIT FOR NETWORK {} SECS".format(WAIT_INIT))
    args = p.parse_args()
    _verbose = not args.silent

    logging.basicConfig(filename=LOG_FILE, level=LOG_LEVEL,
                        datefmt='%d/%m/%Y %H:%M:%S',
                        format='%(levelname)s [%(filename)s_%(funcName)s] - %(asctime)s: %(message)s')

    if not args.nowait:
        time.sleep(WAIT_INIT)
    path_home = '/home/homeassistant'

    os.chdir(path_home)
    logging.debug('INIT CRONIP con path_home: {}'.format(path_home))

    if args.print:
        with open(LOG_FILE) as f:
            log_lines = f.readlines()
        print('LAST {} LOG ENTRIES:'.format(50))
        [print(l[:-1]) for l in log_lines[-50:]]
    elif args.delete:
        logging.info('LOG DELETED (SIZE=0) --> {}'.format(LOG_FILE))
        print('LOG DELETED (SIZE=0) --> {}'.format(LOG_FILE))
        with open(LOG_FILE, 'w') as f:
            f.write('')
    else:
        results = get_machine_info(_verbose)
        results.update({'path_home': path_home})
        ok_ip_ext, ext_ip, location_info, time_ip_ext = get_ext_ip()
        msg = 'RESULTADOS:\n{}\n--> IP_EXT: {} (OK: {}, TIME: {} seg), INFO: {}'
        msg = msg.format(results, ext_ip, ok_ip_ext, time_ip_ext, location_info)
        if _verbose:
            print(msg)
        logging.debug(msg)
        results.update({'hay_ip_ext': ok_ip_ext, 'ip_ext': ext_ip,
                        'time_ip_ext': time_ip_ext,
                        'location_info': location_info})
        if ok_ip_ext:
            _notify_email(results, args.extra, args.email)
        else:
            second_retry = 0
            while second_retry < 2 and not ok_ip_ext:
                logging.error(
                    'NO SE ENCUENTRA IP EXTERNA. IP: {}'.format(results['IP']))
                second_retry += 1
                time.sleep(TIME_PAUSE)
                ok_ip_ext, ext_ip, location_info, time_ip_ext = get_ext_ip(
                    timeout=15)
                results.update({'hay_ip_ext': ok_ip_ext, 'ip_ext': ext_ip,
                                'time_ip_ext': time_ip_ext})
            _notify_email(results, args.extra, args.email)
            logging.info('ENCONTRADA IP EXTERNA. IP: {}, INFO: {}'.format(
                results['ip_ext'], location_info))
        if ok_ip_ext:  # check HA location config
            time.sleep(WAIT_INIT / 2)
            _check_location(location_info, not args.silent)



# -*- coding: utf-8 -*-
"""
Flask routes para facturación del consumo eléctrico en España.

"""
import datetime as dt
from flask import Response, render_template, url_for, request
import json
from time import time

from esiosdata.facturapvpc import FacturaElec, DATOS_ZONAS_IMPUESTOS, DATOS_TIPO_PEAJE
from enerpi.base import log, SENSORS, BILLING_DATA
from enerpi.api import enerpi_data_catalog
from enerpiweb import app, auto


def _format_event_stream(d_msg):
    return 'data: {}\n\n'.format(json.dumps(d_msg))


def _gen_stream_data_factura(start=None, end=None, **kwargs_factura):
    tic = time()
    if start is None:
        # Inicio de mes actual hasta instante actual
        start = dt.datetime.now(tz=SENSORS.TZ).replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        end = dt.datetime.now(tz=SENSORS.TZ)
    cat = enerpi_data_catalog(check_integrity=False)
    df = cat.get_summary(start=start, end=end)
    toc_df = time()
    if (df is not None) and not df.empty and ('kWh' in df):
        consumption = df['kWh']
        # Fix timezone (ya en esiosdata)
        # try:
        #     consumption.index = consumption.index.tz_localize(SENSORS.TZ, ambiguous='infer')
        # except AmbiguousTimeError as e:
        #     consumption.index = consumption.index.tz_localize(SENSORS.TZ, ambiguous='NaT')
        #     consumption = consumption.reindex(DatetimeIndex(start=consumption.index[0], end=consumption.index[-1],
        #                                                     freq='1h', tz=SENSORS.TZ)).interpolate()
        #     log('AmbiguousTimeError ({}) en elec_bill. Se reindexa e interpola el índice.'.format(e), 'error')
        factura = FacturaElec(consumo=consumption, **kwargs_factura)
        data_factura = factura.to_dict(include_text_repr=True, include_html_repr=True)
        toc_p = time()
        msg = 'Factura generada en {:.3f} s; datos en {:.3f} s.'.format(toc_p - toc_df, toc_df - tic)
        log(msg, 'debug', False)
        log('stream_data_factura: STREAM BILL from "{}" to "{}"'.format(start, end), 'debug', False)
        yield _format_event_stream(dict(success=True, factura=data_factura,
                                        took=round(toc_p - tic, 3), took_df=round(toc_df - tic, 3)))
    elif df is not None:
        factura = FacturaElec(start, end, **kwargs_factura)
        data_factura = factura.to_dict(include_text_repr=True, include_html_repr=True)
        toc_p = time()
        msg = 'Factura vacía generada en {:.3f} s; datos en {:.3f} s.'.format(toc_p - toc_df, toc_df - tic)
        log(msg, 'debug', False)
        log('stream_data_factura: STREAM EMPTY BILL from "{}" to "{}"'.format(start, end), 'debug', False)
        yield _format_event_stream(dict(success=True, factura=data_factura, error=msg,
                                        took=round(toc_p - tic, 3), took_df=round(toc_df - tic, 3)))
    else:
        msg = 'No data from {} to {}. CATALOG:\n{}'.format(start, end, cat)
        log(msg, 'debug', False)
        log('stream_data_factura: STREAM ERR NO DATA from "{}" to "{}"'.format(start, end), 'debug', False)
        yield _format_event_stream(dict(success=False, error=msg,
                                        took=round(time() - tic, 3), took_df=round(toc_df - tic, 3)))
    log('CLOSING stream_data_factura from "{}" to "{}" with args={}'.format(start, end, kwargs_factura), 'debug', False)
    yield _format_event_stream('CLOSE')


###############################
# BILLING DATA
###############################
@app.route('/api/billing', methods=['GET'])
@app.route('/api/billing/from/<start>', methods=['GET'])
@app.route('/api/billing/from/<start>/to/<end>', methods=['GET'])
@auto.doc()
def billing_data(start=None, end=None):
    """
    Stream the billing data to make a report.
    Used for load/reload report from user queries.

    :param start: :str: start datetime of data report
    :param end: :str: end datetime of data report

    :return: SSE stream response

    """
    kwargs = dict(start=start, end=end, cups=BILLING_DATA['cups'])
    if 'potencia' in request.args:
        kwargs.update(potencia_contratada=float(request.args['potencia']))
    if 'bono_social' in request.args:
        kwargs.update(con_bono_social=request.args['bono_social'].lower() == 'true')
    if 'impuestos' in request.args:
        zonas = list(DATOS_ZONAS_IMPUESTOS)
        kwargs.update(zona_impuestos=zonas[int(request.args['impuestos'])])
    if 'peaje' in request.args:
        peajes = list(DATOS_TIPO_PEAJE)
        kwargs.update(tipo_peaje=peajes[int(request.args['peaje'])])
    log('BILLING_DATA: {}'.format(kwargs), 'debug', False)
    return Response(_gen_stream_data_factura(**kwargs), mimetype='text/event-stream')


#############################
# BILLING
#############################
@app.route('/api/bills', methods=['GET'])
@auto.doc()
def elec_bill():
    """Endpoint for get the electrical bill (SPAIN Only, PVPC model) of a consumption period."""
    # Start with the current month:
    now = dt.datetime.now()
    ts_ini = '{:%Y-%m-%d}'.format(now.replace(day=1).date())
    ts_fin = '{:%Y-%m-%d}'.format(now.date())
    url_factura_init = url_for('billing_data', peaje=BILLING_DATA['peaje'], impuestos=BILLING_DATA['zona_impuestos'],
                               bono_social=BILLING_DATA['con_bono'], potencia=BILLING_DATA['p_contrato'], )
    log('In base page for BILLING_DATA with init bill in {}'.format(url_factura_init), 'debug', False)
    return render_template('elec_bill.html', url_factura_init=url_factura_init, ts_ini=ts_ini, ts_fin=ts_fin,
                           zonas_impuestos=DATOS_ZONAS_IMPUESTOS, tipos_peaje=DATOS_TIPO_PEAJE, **BILLING_DATA)



# -*- coding: utf-8 -*-
"""
ENERPI - python API for accessing handy objects:
- Remote replication or data extraction
- Local configuration variables
- Local ENERPI data catalog
- Constant & config derived definitions
- Receiver value generator
...

"""
import datetime as dt
import os
import pandas as pd
import requests
import tempfile
# noinspection PyUnresolvedReferences
from enerpi.base import (ENCODING, CONFIG, SENSORS, DATA_PATH, INDEX_DATA_CATALOG, check_resource_files,
                         FILE_LOGGING, LOGGING_LEVEL, log, timeit)
# noinspection PyUnresolvedReferences
from enerpi.database import init_catalog, get_ts_last_save, extract_log_file, delete_log_file, HDF_STORE
# noinspection PyUnresolvedReferences
from enerpi.enerpimeter import receiver_msg_generator, enerpi_raw_data, msg_to_dict
# noinspection PyUnresolvedReferences
from enerpi.iobroadcast import get_encryption_key, get_codec
from enerpi.hdftscat import get_catalog_paths


def enerpi_receiver_generator(verbose=False, n_msgs=None):
    """
    Generator of broadcasted values by ENERPI Logger.

    It can be used by any machine in the same network as the ENERPI Logger. It decrypts the encrypted broadcast and
    returns a dict of vars - values.
    Used by the webserver for read & stream real-time values.

    :param verbose: :bool: Log to stdout
    :param n_msgs: :int: # of msgs to receive (unlimited by default).
    :return: :dict:
    """
    gen = receiver_msg_generator(verbose=verbose, n_msgs=n_msgs)
    count = 0
    while True:
        try:
            msg, _t1, _t2 = next(gen)
            yield msg_to_dict(msg)
            count += 1
        except StopIteration:
            log('EXIT from enerpi_receiver_generator. StopIteration in msg #{}'.format(count), 'error', verbose)
            break
    return None


def enerpi_default_config():
    """
    Default configuration for ENERPI Data Catalog, read from INI file.

    :return: :dict: parameters
    """
    conf = {'store': HDF_STORE,
            'DATA_PATH': DATA_PATH,
            'delta': SENSORS.delta_sec_data,
            'window': SENSORS.rms_roll_window_sec,
            'ts': SENSORS.ts_data_ms,
            'LOGGING_LEVEL': LOGGING_LEVEL,
            'FILE_LOGGING': FILE_LOGGING}
    return conf


def enerpi_data_catalog(check_integrity=False, **kwargs):
    """
    Get ENERPI data catalog for access & operation.

    :param check_integrity: :bool: False by default. If true, checks integrity and generates / updates data index.
    :param kwargs: :dict:
    :return: :EnerpiCatalog:
    """
    return init_catalog(check_integrity=check_integrity, **kwargs)


def _request_enerpi_remote_file(url_api_download, dest_path,
                                file_mimetype='application/octet-stream', verbose=False, timeout=10):
    """
    Download remote file from ENERPI

    :param str url_api_download: download URL
    :param str dest_path: local path for remote file
    :param str file_mimetype: mimetype check
    :param bool verbose: verbose mode
    :return: download ok
    :rtype: bool

    """
    try:
        r = requests.get(url_api_download, timeout=timeout)
    except requests.exceptions.ConnectTimeout:
        log('TIMEOUT REQUEST AT "{}"'.format(url_api_download), 'error', verbose, False)
        return False
    if r.ok and (r.headers['Content-Type'] == file_mimetype):
        if verbose:
            size_kb = int(r.headers['Content-Length']) / 1024
            date_mod = pd.Timestamp(r.headers['Last-Modified']).tz_convert('Europe/Madrid')
            log('ENERPI FILE Downloaded from "{}" --> {:.2f} KB, mtime={:%d/%m/%Y %H:%M:%S}'
                .format(url_api_download, size_kb, date_mod), 'ok', verbose, False)
        check_resource_files(dest_path, verbose=False)
        with open(dest_path, 'wb') as f:
            f.write(r.content)
        if verbose:
            local_size = os.path.getsize(dest_path) / 1024
            local_date_mod = dt.datetime.fromtimestamp(os.path.getmtime(dest_path))
            log('DOWNLOADED FILE NOW IN LOCAL DISK AT "{}", {:.2f} KB, mtime={:%d/%m/%Y %H:%M:%S}'
                .format(dest_path, local_size, local_date_mod), 'magenta', verbose, False)
        return True
    log('REQUEST NOT OK TRYING TO DOWNLOAD FILE AT "{}". STATUS_CODE={}, HEADERS={}'
        .format(url_api_download, r.status_code, r.headers), 'error', verbose, False)
    return False


@timeit('remote_data_get', verbose=True)
def remote_data_get(t0, tf=None,
                    enerpi_ip='192.168.1.52', port=80,
                    prefix_remote_enerpi='/enerpi', verbose=True):
    """
    Query a ENERPI catalog in a remote machine with enerpiweb running.

    :param t0: start of slice
    :param tf:  end of slice (or None for end = now)
    :param enerpi_ip: IP of the remote machine
    :param port: PORT of the remote enerpiweb server
    :param prefix_remote_enerpi: URL prefix of the remote enerpiweb server
    :param verbose: :bool: verbose mode
    :return: (data_key, pd.DataFrame of sliced values) pairs
    :rtype: dict

    """
    def _request_extract_enerpi_data_store(url_api_download_st):
        dest_path = os.path.join(tempfile.gettempdir(), 'temp_store.h5')
        'application/octet-stream'
        data = _request_enerpi_remote_file(url_api_download_st, dest_path, file_mimetype='application/octet-stream')
        if data:
            # Return content & remove hdf temporal file store:
            with pd.HDFStore(dest_path, 'r') as st:
                data = {k: st[k] for k in st.keys()}
                log('HDF Store downloaded:\n{}'.format(st), 'ok', verbose, False)
            os.remove(dest_path)
            return data
        return None

    paths = get_catalog_paths(t0, tf)
    url_mask = 'http://{}:{}{}/api/hdfstores/'.format(enerpi_ip, port, prefix_remote_enerpi) + '{}'
    data_stores = []
    for p in paths:
        url = url_mask.format(os.path.split(p)[-1])
        log('REQUEST HDF STORE AT: {}'.format(url), 'info', verbose, False)
        data_i = _request_extract_enerpi_data_store(url)
        if data_i is not None:
            data_stores.append(data_i)
    keys = set([k for d in data_stores for k in d.keys()])
    data_out = {k: pd.DataFrame(pd.concat([data[k].loc[t0:tf] for data in data_stores
                                           if k in data.keys()])).sort_index()
                for k in keys}
    return data_out


@timeit('replicate_remote_enerpi_data_catalog', verbose=True)
def replicate_remote_enerpi_data_catalog(local_path=DATA_PATH, enerpi_ip='192.168.1.52', port=80,
                                         prefix_remote_enerpi='/enerpi', verbose=True):
    """
    Replicate the ENERPI data catalog from a remote machine with enerpiweb running.

    :param str local_path: Local path where to replicate remote data catalog
    :param str enerpi_ip: IP of the remote machine
    :param int port: PORT of the remote enerpiweb server (Default 80)
    :param prefix_remote_enerpi: URL prefix of the remote enerpiweb server (Default '/enerpi')
    :param bool verbose: verbose mode

    """
    url_mask = 'http://{}:{}{}/api/'.format(enerpi_ip, port, prefix_remote_enerpi) + '{}'
    # Get remote index:
    csv_url = url_mask.format('filedownload/catalog')
    path_csv = os.path.join(local_path, INDEX_DATA_CATALOG)
    mimetype_csv = 'text/csv; charset={}'.format(ENCODING.lower())
    ok_catalog = _request_enerpi_remote_file(csv_url, path_csv, file_mimetype=mimetype_csv, verbose=verbose, timeout=60)
    if not ok_catalog:
        log('ERROR RETRIEVING REMOTE CATALOG FILE IN: {}, ok={}'.format(csv_url, ok_catalog), 'error', verbose, False)
        return False
    # Load retrieved catalog:
    remote_cat = enerpi_data_catalog(check_integrity=False, base_path=local_path)
    if (remote_cat.tree is None) or remote_cat.tree.empty:
        log('EMPTY REMOTE CATALOG! NOTHING TO REPLICATE. EXITING...', 'error', verbose, False)
        return False
    else:
        log('REMOTE CATALOG TO REPLICATE:\n{}'.format(remote_cat.tree), 'debug', verbose, False)
        df_stores = remote_cat.tree[remote_cat.tree.is_cat & remote_cat.tree.is_raw]
        result = {}
        ts_init = df_stores.ts_ini.min()
        for _, row in df_stores.iterrows():
            rel_path_remote_st = row.st
            log('* Replicating store "{}", with {} raw samples from {:%-d/%m/%Y} to {:%-d/%m/%Y}'
                .format(rel_path_remote_st, row.n_rows, row.ts_ini, row.ts_fin), 'debug', verbose, False)
            abs_path_new_st = os.path.join(local_path, rel_path_remote_st)
            file_id = os.path.split(rel_path_remote_st)[-1]
            url_remote_st = url_mask.format('hdfstores/' + file_id)
            ok_store_i = _request_enerpi_remote_file(url_remote_st, abs_path_new_st,
                                                     file_mimetype='application/octet-stream', verbose=verbose)
            if ok_store_i:
                result[file_id] = (int(os.path.getsize(abs_path_new_st) / 1024), row.n_rows)
            else:
                result[file_id] = (None, row.n_rows)
        # Download raw_data:
        url_remote_raw_st = url_mask.format('filedownload/raw_store')
        abs_path_new_raw_data = os.path.join(local_path, HDF_STORE)
        ok_store_raw = _request_enerpi_remote_file(url_remote_raw_st, abs_path_new_raw_data,
                                                   file_mimetype='application/octet-stream', verbose=verbose)
        # Operation report
        msg = '\nREPLICATION FROM {} DONE. DATA SINCE: {:%c}. RESULTS:\n'.format(enerpi_ip, ts_init)
        kbytes = 0
        if ok_store_raw:
            kbytes_i = int(os.path.getsize(abs_path_new_raw_data) / 1024)
            kbytes += kbytes_i
            msg += '  -> RAW DATA: {} KB\n'.format(kbytes_i)
        for k, v in result.items():
            kbytes += v[0]
            msg += '  -> STORE "{}": {} KB; {} ROWS\n'.format(k, v[0], v[1])
        msg += 'TOTAL SIZE OF DOWNLOADED DATA: {:.2f} MB'.format(kbytes / 1024)
        log(msg, 'info', verbose, False)
        return True



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import bokeh
from bokeh.plotting import figure
from bokeh.embed import components
from bokeh.models import HoverTool, ColumnDataSource, LinearAxis, Range1d
from bokeh.io import reset_output
from enerpi.base import SENSORS
from enerpiplot import (ROUND_W, ROUND_KWH, COLS_DATA_KWH, COLORS_DATA_KWH, COLOR_REF_RMS,
                        UNITS_DATA_KWH, LABELS_DATA_KWH, FMT_TOOLTIP_DATA_KWH)


COL_TS = SENSORS.ts_column

# TOOLS = "pan,xwheel_zoom,ywheel_zoom,box_zoom,reset,save,crosshair"
TOOLS = "pan,xwheel_zoom,box_zoom,reset,save,crosshair"
P_WIDTH = 900
P_HEIGHT = 500

HTML_TROW = """<tr><td style="font-size: 15px; font-weight: bold;">
<span class="bk-tooltip-color-block" style="background-color: {2}"></span>{0}</td>
<td style="font-size: 16px; font-weight: bold; color: {2};">{1}</td></tr>"""
TOOLTIP_ROWS = """<tr><td style="font-size: 15px; font-weight: bold;">Time:</td>
<td style="font-size: 15px; font-weight: bold;">@time</td></tr>"""


def _append_hover(df, delta_min=0, multi_day=False):
    if multi_day:
        df['time'] = [(x + pd.Timedelta(minutes=delta_min)).strftime("%-d/%b %H:%M:%S") for x in df.index]
    else:
        df['time'] = [(x + pd.Timedelta(minutes=delta_min)).strftime("%H:%M:%S.%f")[:-3] for x in df.index]
    return df


def _titulo_rango_temporal(extremos_dia):
    n_days = (extremos_dia[1] - extremos_dia[0]).days
    if n_days > 1:
        return "{:%-d de %B'%y} --> {:%-d de %B'%y}".format(extremos_dia[0], extremos_dia[1])
    else:
        return '{:%A %-d de %B de %Y}'.format(extremos_dia[0])


def _get_extremos_dia(df):
    extremos = [x.replace(microsecond=0) for x in df.index[[0, -1]]]
    # extremos_dia = [pd.Timestamp(x.date() + pd.Timedelta(days=x.hour // 12)) for x in extremos]
    # return [extremos[0].date(), (extremos[1] - pd.Timedelta(minutes=1) + pd.Timedelta(days=1)).date()]
    return extremos


def get_bokeh_version():
    """
    For jinja2 templating (css & js src's)
    """
    return bokeh.__version__


def _return_html_comps(plots):
    script, divs = components(plots)
    reset_output()
    return script, divs, get_bokeh_version()


def _get_figure_plot(extremos, y_range):
    kwargs = dict(x_range=Range1d(*extremos), y_range=y_range,
                  tools=TOOLS, active_drag=None, plot_width=P_WIDTH, plot_height=P_HEIGHT,
                  x_axis_type="datetime", toolbar_location="right", toolbar_sticky=False,  # "above",
                  title=_titulo_rango_temporal(extremos), responsive=True)
    return figure(**kwargs)


def _format_axis_plot(p, color, label, unit):
    # Axis formatting
    p.axis.major_tick_in = 7
    p.axis.minor_tick_in = 3
    p.axis.major_tick_out = p.axis.minor_tick_out = 0
    p.title.text_font_size = '12pt'
    if unit:
        p.yaxis.axis_label = '{} ({})'.format(label, unit)
    else:
        p.yaxis.axis_label = label
    p.yaxis.axis_label_text_font_size = "11pt"
    p.yaxis.axis_line_color = p.yaxis.major_label_text_color = p.yaxis.axis_label_text_color = color
    p.yaxis.major_tick_line_color = p.yaxis.minor_tick_line_color = color


def _format_legend_plot(p):
    # Legend formatting
    p.legend.location = "top_left"
    p.legend.label_text_font_size = "7pt"
    p.legend.background_fill_alpha = .5
    p.legend.label_height = 12
    p.legend.glyph_height = 10
    p.legend.label_standoff = 5
    p.legend.padding = 8
    p.legend.spacing = 3


def _plot_patch_series(fig_plot, pd_series, **kwargs_patch):
    zero = np.array([0])
    df_patch = pd.Series(pd_series.fillna(method='pad', limit=2).fillna(0).round(2))
    x = np.append(np.insert(df_patch.index.values, 0, df_patch.index.values[0]), df_patch.index.values[-1])
    y = np.append(np.insert(df_patch.values, 0, zero), zero)
    fig_plot.patch(x, y, **kwargs_patch)


def _get_columns_data_for_raw_plot(columns=None):
    col_main_rms = col_main_mean = None
    cols_sensors_rms = SENSORS.columns_sensors_rms
    cols_sensors_mean = SENSORS.columns_sensors_mean
    if columns is not None:
        cols_sensors_rms = list(filter(lambda x: x in columns, cols_sensors_rms))
        cols_sensors_mean = list(filter(lambda x: x in columns, cols_sensors_mean))
    if cols_sensors_rms:
        col_main_rms = cols_sensors_rms[0]
    if cols_sensors_mean:
        col_main_mean = cols_sensors_mean[0]
    cols_data = cols_sensors_rms + cols_sensors_mean
    b_sensors_rms = [True] * len(cols_sensors_rms) + [False] * len(cols_sensors_mean)
    colors_data = [SENSORS[s].color for s in cols_data]
    units_data = [SENSORS[s].unit for s in cols_data]
    fmt_tooltip_data = ['{0}'] * len(cols_sensors_rms) + ['{0.0}'] * len(cols_sensors_mean)
    if (columns is None) or (SENSORS.ref_rms in columns):
        cols_data += [SENSORS.ref_rms]
        colors_data += [COLOR_REF_RMS]
        units_data += ['']
        fmt_tooltip_data += ['{0}']
        b_sensors_rms += [True]
    return {"main_rms": col_main_rms, "main_mean": col_main_mean,
            "sensors_rms": cols_sensors_rms, "sensors_mean": cols_sensors_mean,
            "labels": SENSORS.descriptions(cols_data), "colsdata": cols_data, "colors": colors_data,
            "units": units_data, "tooltip_fmts": fmt_tooltip_data, "b_sensors": b_sensors_rms}


def _plot_bokeh_raw(data_plot, columns=None):
    """Bokeh plot de datos en bruto."""
    # Plot params:
    confp = _get_columns_data_for_raw_plot(columns)
    # Bokeh does not work very well!! with timezones:
    data_plot = data_plot.tz_localize(None)
    for c in confp["sensors_mean"]:
        # Corrección a % de 0 a 100
        data_plot[c] /= 10.
    if confp["main_rms"]:
        y_range = [0, max(500, int(np.ceil(data_plot[[confp["main_rms"],
                                                      SENSORS.ref_rms]].max().max() / ROUND_W) * ROUND_W))]
        minmax_ejes = [y_range, [0, 100]]
    else:
        minmax_ejes = [[0, 100]]
        # Elimina columna de # samples si no hay RMS columns
        if SENSORS.ref_rms in confp["colsdata"]:
            confp["colsdata"].remove(SENSORS.ref_rms)

    extremos = _get_extremos_dia(data_plot)
    ejes = [confp["main_rms"]]
    if confp["main_mean"]:
        ejes += [confp["main_mean"]]
    pos_smean = len(confp["sensors_rms"])

    # Figure
    p = _get_figure_plot(extremos, minmax_ejes[0])

    zip_props = zip(confp["colsdata"], confp["colors"], confp["units"], confp["labels"], confp["tooltip_fmts"])
    tooltip_rows = TOOLTIP_ROWS
    for c, color, unit, label, fmt in zip_props:
        tooltip_rows += HTML_TROW.format('{}:'.format(label), '@{}{} {}'.format(c, fmt, unit), color)
    p.add_tools(HoverTool(tooltips='<div><table>{}</table></div>'.format(tooltip_rows)))

    # Axis formatting
    _format_axis_plot(p, confp["colors"][0], confp["labels"][0], confp["units"][0])
    if len(ejes) > 1:
        positions = (['right', 'left'] * 2)[:len(ejes[1:])]
        zip_ejes = zip(ejes[1:], positions, minmax_ejes[1:], confp["colors"][pos_smean:],
                       confp["labels"][pos_smean:], confp["units"][pos_smean:])
        for extra_eje, pos, minmax, color, label, unit in zip_ejes:
            p.extra_y_ranges[extra_eje] = Range1d(*minmax)
            axkw = dict(y_range_name=extra_eje,
                        axis_label='{} ({})'.format(label, unit),
                        axis_label_text_font_size="10pt", axis_line_color=color,
                        major_label_text_color=color, axis_label_text_color=color,
                        major_tick_line_color=color, minor_tick_line_color=color,
                        major_tick_in=5, major_tick_out=0, minor_tick_in=3, minor_tick_out=0)
            p.add_layout(LinearAxis(**axkw), pos)

    # Make data source w/ time hover
    data = ColumnDataSource(_append_hover(data_plot, multi_day=(extremos[1] - extremos[0]).days > 1))

    # Plot lines of analog sensors
    for s, is_rms, color in zip(confp["colsdata"], confp["b_sensors"], confp["colors"]):
        kwargs_p = dict(source=data, alpha=.9, line_width=2, color=color)
        if confp["main_rms"] and not is_rms:
            kwargs_p.update(dict(y_range_name=ejes[1]))
        p.line(COL_TS, s, **kwargs_p)

    # Plot patch of main rms & normal analog sensors
    if confp["main_mean"]:
        kwargs_patch = dict(color=confp["colors"][pos_smean], line_alpha=0, fill_alpha=0.10)
        if confp["main_rms"]:
            kwargs_patch.update({"y_range_name": ejes[1]})
        _plot_patch_series(p, data_plot[confp["main_mean"]], **kwargs_patch)

    if confp["main_rms"]:
        kwargs_patch = dict(color=confp["colors"][0], line_alpha=0, fill_alpha=0.15)
        _plot_patch_series(p, data_plot[confp["main_rms"]], **kwargs_patch)

    # Legend formatting
    # if len(confp["colsdata"]) > 1:
    #     _format_legend_plot(p)
    return p


# TODO Arreglar summary plot añadiendo precio con ejes kwh/€/kW_max_min
def _plot_bokeh_hourly(data_plot, columns=None):
    # Bokeh does not work very well!! with timezones:
    data_plot = data_plot.tz_localize(None)

    data_plot['ts_mid'] = data_plot.index + pd.Timedelta('30min')
    data_plot.index += pd.Timedelta('30min')

    y_range = [0, max(1, np.ceil(data_plot[COLS_DATA_KWH[0]].max().max() / ROUND_KWH) * ROUND_KWH)]
    minmax_ejes = [y_range, [0, max(500, int(np.ceil(data_plot[COLS_DATA_KWH[1]].max().max() / ROUND_W) * ROUND_W))]]
    extremos = _get_extremos_dia(data_plot)
    ejes = COLS_DATA_KWH[:2]

    # Figure
    p = _get_figure_plot(extremos, minmax_ejes[0])

    tooltip_rows = TOOLTIP_ROWS
    for c, color, unit, label, fmt in zip(COLS_DATA_KWH, COLORS_DATA_KWH,
                                          UNITS_DATA_KWH, LABELS_DATA_KWH, FMT_TOOLTIP_DATA_KWH):
        tooltip_rows += HTML_TROW.format('{}:'.format(label), '@{}{} {}'.format(c, fmt, unit), color)
    p.add_tools(HoverTool(tooltips='<div><table>{}</table></div>'.format(tooltip_rows)))

    # Axis formatting
    _format_axis_plot(p, COLORS_DATA_KWH[0], LABELS_DATA_KWH[0], UNITS_DATA_KWH[0])
    if len(ejes) > 1:
        positions = (['right', 'left'] * 2)[:len(ejes[1:])]
        for extra_eje, pos, minmax, color, label, unit in zip(ejes[1:], positions, minmax_ejes[1:], COLORS_DATA_KWH[1:],
                                                              LABELS_DATA_KWH[1:], UNITS_DATA_KWH[1:]):
            p.extra_y_ranges[extra_eje] = Range1d(*minmax)
            axkw = dict(y_range_name=extra_eje,
                        axis_label='{} ({})'.format(label, unit),
                        axis_label_text_font_size="10pt", axis_line_color=color,
                        major_label_text_color=color, axis_label_text_color=color,
                        major_tick_line_color=color, minor_tick_line_color=color,
                        major_tick_in=5, major_tick_out=0, minor_tick_in=3, minor_tick_out=0)
            p.add_layout(LinearAxis(**axkw), pos)

    # Make data source w/ time hover
    data = ColumnDataSource(_append_hover(data_plot, delta_min=-30, multi_day=(extremos[1] - extremos[0]).days > 1))

    # Plot lines
    if (columns is None) or (COLS_DATA_KWH[0] in columns):
        kwargs_kwh = dict(source=data, x=COL_TS, width=3600000, bottom=0, top=COLS_DATA_KWH[0],
                          legend=LABELS_DATA_KWH[0], fill_alpha=.7, line_alpha=.9, line_width=1,
                          line_join='round', color=COLORS_DATA_KWH[0])
        p.vbar(**kwargs_kwh)
    if (columns is None) or (COLS_DATA_KWH[1] in columns):
        kwargs_l = dict(source=data, alpha=.95, line_width=1.5, line_join='bevel', color=COLORS_DATA_KWH[1],
                        y_range_name=COLS_DATA_KWH[1], legend=LABELS_DATA_KWH[1])
        p.line(COL_TS, COLS_DATA_KWH[1], **kwargs_l)
    if (columns is None) or (COLS_DATA_KWH[2] in columns):
        kwargs_l = dict(source=data, alpha=.95, line_width=1.5, line_join='bevel', color=COLORS_DATA_KWH[2],
                        y_range_name=COLS_DATA_KWH[1], legend=LABELS_DATA_KWH[2])
        p.line(COL_TS, COLS_DATA_KWH[2], **kwargs_l)

    # Legend formatting
    _format_legend_plot(p)
    return p


def html_plot_buffer_bokeh(data_plot, is_kwh_plot=False, columns=None):
    """
    Given a pandas DataFrame (or a list of df's), returns the html components for rendering the graph.
    :return script, divs, bokeh.__version__
    """
    if is_kwh_plot:
        return _return_html_comps([_plot_bokeh_hourly(data_plot, columns)])
    else:
        return _return_html_comps([_plot_bokeh_raw(data_plot, columns)])



# -*- coding: utf-8 -*-
"""
ENERPIWEB - Flask web application

"""
from flask import Flask
from flask_autodoc import Autodoc
from flask_mail import Mail
from werkzeug.contrib.fixers import ProxyFix
from werkzeug.routing import Rule
import jinja2
import os
# noinspection PyUnresolvedReferences
from enerpi import __version__
from enerpi.base import (CONFIG, DATA_PATH, check_resource_files,
                         GMAIL_ACCOUNT, GMAIL_APP_PASSWORD)
from enerpi.api import get_encryption_key


basedir = os.path.abspath(os.path.dirname(__file__))
STATIC_PATH = os.path.join(
    DATA_PATH, CONFIG.get('ENERPI_WEBSERVER', 'STATIC_PATH', fallback='WWW'))
LOGGING_LEVEL_SERVER = CONFIG.get(
    'ENERPI_WEBSERVER', 'LOGGING_LEVEL_WEB', fallback='DEBUG')
SERVER_FILE_LOGGING = os.path.join(
    STATIC_PATH, CONFIG.get(
        'ENERPI_WEBSERVER', 'FILE_LOGGING_WEB', fallback='enerpiweb.log'))
PREFIX_WEB = CONFIG.get('ENERPI_WEBSERVER', 'PREFIX_WEB', fallback='/enerpi')
BASECOLOR = '#{}'.format(
    CONFIG.get('ENERPI_WEBSERVER', 'BASECOLOR_HEX', fallback='0CBB43'))
check_resource_files(
    STATIC_PATH, os.path.join(basedir, 'static'), verbose=False)

# WITH_WEB = CONFIG.getboolean('ENERPI_WEBSERVER', 'WITH_WEBSERVER',
# fallback=True)
WITH_ML_SUBSYSTEM = CONFIG.getboolean('ENERPI_WEBSERVER', 'WITH_ML',
                                      fallback=False)

# Shows more info (timming data, debug data, ...)
EXTRA_INFO = CONFIG.getboolean('ENERPI_WEBSERVER', 'EXTRA_INFO',
                               fallback=False)

# FLASK APP
app = Flask(__name__, static_path=PREFIX_WEB + '/static',
            static_folder=STATIC_PATH)
app.url_rule_class = lambda path, **options: Rule(PREFIX_WEB + path, **options)
app.jinja_env.trim_blocks = True
app.jinja_env.lstrip_blocks = True
app.jinja_env.cache = {}
app.jinja_loader = jinja2.FileSystemLoader(os.path.join(basedir, 'templates'))

# Manual activation of test-mode
# app.config['TESTING'] = True
# app.config['PROPAGATE_EXCEPTIONS'] = True

# Forms protection
app.config['CSRF_ENABLED'] = True
app.config['WTF_CSRF_ENABLED'] = True
app.config['SECRET_KEY'] = get_encryption_key()

app.config['STREAM_MAX_TIME'] = 1800
app.config['BASECOLOR'] = BASECOLOR
app.config['WITH_ML_SUBSYSTEM'] = WITH_ML_SUBSYSTEM
app.config['EXTRA_INFO'] = EXTRA_INFO

# Plug-ins: email server
if GMAIL_ACCOUNT and GMAIL_APP_PASSWORD:
    app.config['MAIL_SERVER'] = 'smtp.gmail.com'
    app.config['MAIL_PORT'] = 587  # 465
    app.config['MAIL_USE_TLS'] = True
    app.config['MAIL_USE_TLS'] = True
    app.config['MAIL_USE_SSL'] = False
    app.config['MAIL_USERNAME'] = GMAIL_ACCOUNT
    app.config['MAIL_DEFAULT_SENDER'] = GMAIL_ACCOUNT
    app.config['MAIL_PASSWORD'] = GMAIL_APP_PASSWORD
    mail = Mail(app)
else:
    mail = None
# Plug-ins: API Auto-doc:
auto = Autodoc(app)

# Views
# noinspection PyUnresolvedReferences,PyPep8
from enerpiweb import (views, views_filehandler, utils, rt_stream,
                       emailing, facturacionpvpc)

# wsgi
app.wsgi_app = ProxyFix(app.wsgi_app)



# -*- coding: utf-8 -*-
"""
ENERPIWEB __main__
    - Install configuration & CRON task for web resources generation
    - Uninstall CRON task
    - Run Flask server

"""
import os
# noinspection PyUnresolvedReferences
from enerpiweb import app as application
from enerpiweb import SERVER_FILE_LOGGING, LOGGING_LEVEL_SERVER
from enerpiweb.command_enerpiweb import main, check_resource_files
from enerpi.base import set_logging_conf


# logging conf
check_resource_files(os.path.dirname(SERVER_FILE_LOGGING), verbose=True)
set_logging_conf(SERVER_FILE_LOGGING, level=LOGGING_LEVEL_SERVER,
                 verbose=True, with_initial_log=True)


if __name__ == '__main__':
    # Para ejecutar 'a mano'
    main()



# -*- coding: utf-8 -*-
"""
ENERPI PushBullet Notifier

- Push text message (ENERPI errors) or file.
- Push to email recipient (with a pushbullet acount!) or to a pushbullet channel ('enerpi_notifications')

Pushbullet API config:
    * Name:                 ENERPI Current Meter BOT
    * Pushbullet channel:   enerpi_notifications
    * Channel URL:          https://www.pushbullet.com/channel?tag=enerpi_notifications
    * PUSHBULLET_TOKEN:     Token from pushbullet account, it needs to be defined in 'enerpi_config.ini'

"""
from configparser import NoOptionError, NoSectionError
import datetime as dt
from pushbullet import Pushbullet
from pushbullet.errors import InvalidKeyError, PushError, PushbulletError
from enerpi.base import CONFIG, log, async_task, RECIPIENT


CHANNEL = 'enerpi_notifications'
PBOBJ = None


def _get_pb_obj():
    # Lazy load
    global PBOBJ

    if PBOBJ is None:
        try:
            pushbullet_token = CONFIG.get('NOTIFY', 'PUSHBULLET_TOKEN')
            if not pushbullet_token:
                raise InvalidKeyError
            PBOBJ = Pushbullet(pushbullet_token)
        except (NoOptionError, NoSectionError, InvalidKeyError) as e:
            log('NO Pushbullet config ({} [{}])'
                .format(e, e.__class__), 'error', False)
            PBOBJ = None
    return PBOBJ


def _send_direct_push_notification(title, content, email=RECIPIENT):
    pb = _get_pb_obj()
    if (pb is not None) and email:
        try:
            pb.push_note(title, content, email=email)
            log('PUSHBULLET NOTIFICATION: {} - {}'
                .format(title, content), 'debug', False)
            return True
        except (PushError, PushbulletError) as exc:
            log('{} ERROR: {}'
                .format(exc.__class__, exc), 'error', True)
    log('NO PUSHBULLET NOTIFICATION (title={}, email={})'
        .format(title, email), 'error', True)
    return False


@async_task
def push_enerpi_error(title_err, msg_error, email=RECIPIENT):
    """
    Send an error message notification.

    Vía pushbullet API, to the default recipient.
    :param title_err: error title
    :param msg_error: error text content
    :param email: target email

    """
    return _send_direct_push_notification(
        'ERROR: {}\n(time:{:%H:%M:%S})'.format(
            title_err, dt.datetime.now()), msg_error, email=email)


@async_task
def push_enerpi_channel_msg(title, msg, channel_tag=CHANNEL):
    """
    Send a text notification (vía pushbullet API)
     to the ENERPI pushbullet channel (timestamped)

    :param title:
    :param msg:
    :param channel_tag: ENERPI CHANNEL

    """
    content = '{}\n(time:{:%H:%M:%S})'.format(msg, dt.datetime.now())
    pb = _get_pb_obj()
    if pb is not None:
        try:
            ch = list(filter(
                lambda x: x.channel_tag == channel_tag, pb.channels))[0]
        except IndexError:
            ch = None
        if ch:
            try:
                ch.push_note(title, content)
                log('PUSHBULLET CHANNEL NOTIFICATION: {} - {}'
                    .format(title, content), 'debug', False)
                return True
            except (PushError, PushbulletError) as exc:
                log('{} ERROR: {}'
                    .format(exc.__class__, exc), 'error', True)
    log('NO PUSHBULLET CHANNEL NOTIFICATION (title={}, msg={}, channel={})'
        .format(title, msg, channel_tag), 'error', True)
    return False



# -*- coding: utf-8 -*-
"""
Flask routes for send emails with Flask-Mail

"""
from flask import redirect, url_for, render_template
from flask_mail import Message
import json
import os
from enerpi.base import (log, DATA_PATH, RECIPIENT, SENSORS,
                         INDEX_DATA_CATALOG, async_task)
from enerpiweb import app, auto, mail, STATIC_PATH, GMAIL_ACCOUNT


#############################
# ENERPI MAILING
#############################
@async_task
def _send_status_email(text_msg, recipients):
    log('SENDING STATUS EMAIL TO {}'.format(recipients), 'debug')
    mask = os.path.join(
        STATIC_PATH, "img/generated/tile_enerpi_data_{}_last_24h.svg")
    data_monitor = SENSORS.to_dict()
    for i, k in enumerate(data_monitor['sensors']):
        try:
            data_monitor['sensors'][i]['tile'] = open(
                mask.format(data_monitor['sensors'][i]['name']), 'r').read()
        except OSError:
            data_monitor['sensors'][i]['tile'] = 'NO TILE'
    try:
        data_monitor['consumption'] = {'tile': open(
            mask.format('kWh'), 'r').read()}
    except OSError:
        data_monitor['consumption'] = {'tile': 'NO kWh TILE'}
    try:
        data_monitor['ref'] = {'tile': open(mask.format('ref'), 'r').read()}
    except OSError:
        data_monitor['ref'] = {'tile': 'NO REF TILE'}
    html_msg = render_template(
        'email/status_email.html', data_monitor=data_monitor, msg=text_msg)

    msg = Message("Status Info",
                  recipients=recipients,
                  sender='"enerPI" <{}>'.format(GMAIL_ACCOUNT),
                  # cc=None,
                  # bcc=GMAIL_ACCOUNT,
                  # attachments=None,
                  reply_to='"enerPI" <{}>'.format(GMAIL_ACCOUNT),
                  # date=None,
                  body=text_msg, html=html_msg)
    if os.path.exists(INDEX_DATA_CATALOG):
        with open(os.path.join(DATA_PATH, INDEX_DATA_CATALOG), 'rb') as fp:
            msg.attach(INDEX_DATA_CATALOG, "text /csv", fp.read())
    if mail is not None:
        mail.send(msg)
        log('EMAIL SENT', 'debug')
    else:
        log('EMAIL NOT SENT --> {}'.format(msg.as_string()), 'error')


@app.route('/api/email/status', methods=['GET'])
@app.route('/api/email/status/<recipients>', methods=['GET'])
@auto.doc()
def send_status_email(recipients=(RECIPIENT,)):
    """GET method for send the enerPI status from last 24 hours
    to the specified email recipients.

    :param recipients: comma separated emails,
    like: '/api/email/status/example@hotmail.com,other@gmail.com'
    :return: Start the sending email process in async mode,
            and redirects to '/control' with an alert.

    """
    if type(recipients) is str:
        recipients = recipients.split(',')
    if recipients[0]:
        text_msg = 'enerPI STATUS from last 24h... '
        _send_status_email(text_msg, recipients)
        alert = json.dumps({'alert_type': 'success',
                            'texto_alerta': ('Sent email to: {}\n->"{}"'
                                             .format(recipients if len(recipients) > 1 else recipients[0], text_msg))})
    else:
        alert = json.dumps({'alert_type': 'danger',
                            'texto_alerta': "Can't send email, no default EMAIL RECIPIENT defined"})
    return redirect(url_for('control', alerta=alert))



# -*- coding: utf-8 -*-
from codecs import open
import os
from setuptools import setup, find_packages
from enerpi import __version__ as version, BASE_PATH


packages = find_packages(exclude=['docs', '*tests*', 'notebooks'])
with open(os.path.join(BASE_PATH, '..', 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='enerpi',
    version=version,
    description='AC Current Meter for Raspberry PI with GPIOZERO and MCP3008',
    long_description='\n' + long_description,
    keywords='enerpi current gpiozero raspberry analog',
    author='Eugenio Panadero',
    author_email='azogue.lab@gmail.com',
    url='https://github.com/azogue/enerpi',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Home Automation',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Natural Language :: Spanish',
        'License :: OSI Approved :: MIT License',
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: Unix'],
    packages=packages,
    package_data={
        'enerpi': ['config/default_config_enerpi.ini',
                   'config/default_sensors_enerpi.json',
                   'config/.enerpi_data_path'],
        'enerpiweb': ['templates/*', 'templates/email/*', 'templates/doc/*',
                      'templates/macros/*',
                      'static/css/*', 'static/img/icons/*', 'static/js/*'],
    },
    install_requires=['numpy>=1.11.2', 'pandas>=0.19.0', 'pytz>=2016.7',
                      'flask>=0.11.1', 'werkzeug>=0.11.11',
                      'Flask-Mail>=0.9.1', 'flask-autodoc>=0.1.2',
                      'Flask-WTF>=0.13.1', 'wtforms>=2.1', 'uwsgi>=2.0.15',
                      'cryptography>=1.5.2',
                      'gpiozero>=1.3.1', 'RPi.GPIO>=0.6.3',
                      'matplotlib>=1.5.3', 'bokeh>=0.12.3', 'termcolor>=1.1.0',
                      'python-crontab>=2.1.1', 'croniter>=0.3.16',
                      'pushbullet.py>=0.10.0', 'jsondiff>=1.0.0',
                      'esiosdata>=0.6.10', 'python-dateutil>=2.6.0',
                      'pillow>=4.0.0'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
    entry_points={
        'console_scripts': [
            'enerpi = enerpi.__main__:enerpi_main_cli',
            'pitemps = enerpi.pitemps.__main__:main',
            'enerpi-daemon = enerpi.enerdaemon:enerpi_daemon',
            'enerpi-rscgen= enerpiplot.mule_rscgen:main',
            'enerpiweb = enerpiweb.__main__:main'
        ]
    },
    # Test configuration:
    # setup_requires=['pytest-runner'],
    tests_require=['pytest>=3.0.0', 'freezegun>=0.3.8'],
    # test_suite='nose.collector',
)



# -*- coding: utf-8 -*-
"""
ENERPI - Database methods:

- Get/Init/Update ENERPI data catalog
- Process or clear log files
- Method for appending raw data to ENERPI catalog from the ENERPI Logger
...

"""
import datetime as dt
import os
import pandas as pd
from time import time
import re
from enerpi.base import CONFIG, DATA_PATH, log, timeit, SENSORS
from enerpi.catalog import EnerpiCatalog


# Config:
INIT_LOG_MARK = CONFIG.get('ENERPI_SAMPLER', 'INIT_LOG_MARK', fallback='INIT ENERPI')
HDF_STORE = CONFIG.get('ENERPI_DATA', 'HDF_STORE')
HDF_STORE_PATH = os.path.join(DATA_PATH, HDF_STORE)

KEY = CONFIG.get('ENERPI_DATA', 'KEY', fallback='/rms')
CONFIG_CATALOG = dict(raw_file=HDF_STORE,
                      key_raw_data=KEY,
                      key_summary_data='/hours',
                      key_summary_extra='/days',
                      check_integrity=True,
                      archive_existent=False,
                      verbose=False)


def init_catalog(sensors=None, base_path=DATA_PATH, **kwargs):
    """
    Get ENERPI data catalog for access & operation with params.

    :param sensors: Sensors config object (class EnerpiSamplerConf)
    :param base_path: :str: ENERPIDATA base path
    :param kwargs: :dict: parameters
    :return: :EnerpiCatalog:
    """
    conf = CONFIG_CATALOG.copy()
    conf.update(base_path=base_path)
    if kwargs:
        conf.update(**kwargs)
    return EnerpiCatalog(sensors=sensors, **conf)


def _clean_store_path(path_st):
    if os.path.pathsep not in path_st:
        path_st = os.path.join(DATA_PATH, path_st)
    else:
        path_st = os.path.abspath(path_st)
    if not os.path.splitext(path_st)[1]:
        path_st += '.h5'
    return path_st


def show_info_data(df, df_consumo=None):
    """
    Prints some info about DATA (& opt SUMMARY_DATA)

    :param df: :pd.DataFrame: DATA
    :param df_consumo: :pd.DataFrame: SUMMARY_DATA
    """
    log('DATAFRAME INFO:\n* Head:\n{}'.format(df.head()), 'info', True, False)
    log('* Tail:\n{}'.format(df.tail()), 'info', True, False)
    log('* Count & types:\n{}'
        .format(pd.concat([df.count().rename('N_rows'), df.dtypes.rename('dtypes'), df.describe().drop('count').T],
                          axis=1)), 'info', True, False)
    if df_consumo is not None and not df_consumo.empty:
        log('\n** HOURLY ELECTRICITY CONSUMPTION (kWh):\n{}'.format(df_consumo), 'magenta', True, False)
        dias = df_consumo.drop(['p_min', 'p_mean', 'p_max'], axis=1).resample('1D').sum()
        p_rs = df_consumo[['p_min', 'p_max']].resample('1D')
        dias = dias.join(p_rs.p_min.min()).join(p_rs.p_max.max())
        dias['t_ref'] /= 24
        log('\n*** DAILY ELECTRICITY CONSUMPTION (kWh):\n{}'.format(dias), 'ok', True, False)


def _notify_error_in_save_raw_data(msg_error):
    from time import sleep
    from enerpi.notifier import push_enerpi_error

    t = push_enerpi_error('SAVE RAW DATA', msg_error)
    sleep(1)
    return t


def save_raw_data(data=None, path_st=HDF_STORE_PATH, catalog=None, verb=True):
    """
    Used in a subprocess launched from enerpimeter, this functions appends new *raw data* to the HDF raw store,
     and, if data-catalog is not None, updates it.
    :param data:
    :param path_st:
    :param catalog:
    :param verb:
    :return:
    """
    try:
        df_tot = None
        if data is not None and type(data) is not pd.DataFrame:
            data = pd.DataFrame(data, columns=SENSORS.columns_sampling
                                ).set_index(SENSORS.ts_column).dropna().astype(float)
            mode = 'a' if os.path.exists(path_st) else 'w'
            try:
                with pd.HDFStore(path_st,
                                 mode=mode, complevel=9, complib='zlib') as st:
                    st.append(KEY, data)
                    if catalog is not None:
                        df_tot = st[KEY]
                        log('Size Store: {:.1f} KB, {} rows'.format(os.path.getsize(path_st) / 1000, len(df_tot)),
                            'debug', verb)
            except OSError as e:
                msg_error = 'OSError "{}" trying to open "{}" in "{}" mode (save_in_store)'.format(e, path_st, mode)
                log(msg_error, 'error', True)
                _notify_error_in_save_raw_data(msg_error)
                return -1
            if catalog is not None:
                try:
                    catalog.update_catalog(data=df_tot)
                except Exception as e:
                    msg_error = 'Exception "{}" [{}] en update_catalog (save_in_store)'.format(e, e.__class__)
                    log(msg_error, 'error', True)
                    _notify_error_in_save_raw_data(msg_error)
                    return -1
        return True
    except ValueError as e:
        log('ValueError en save_in_store: {}'.format(e), 'error', True)
        return -1


@timeit('get_ts_last_save')
def get_ts_last_save(path_st=HDF_STORE_PATH, get_last_sample=False, verbose=True, n=3):
    """
    Returns last data timestamp in hdf store.

    :param path_st: :str: hdf store file path
    :param get_last_sample: :bool: returns ts or pd.DataFrame
    :param verbose: :bool: shows logging msgs in stdout.
    :param n: :int: # of tail rows
    :return: pd.Timestamp or pd.DataFrame
    """
    tic = time()
    try:
        ts = dt.datetime.fromtimestamp(os.path.getmtime(path_st))
        size_kb = os.path.getsize(path_st) / 1000
        if get_last_sample:
            with pd.HDFStore(path_st, mode='r') as st:
                try:
                    df = st.select(KEY, start=-n)
                    log('Store UPDATE: {:%c} , SIZE = {:.2f} KB. TOOK {:.3f} s'.format(ts, size_kb, time() - tic),
                        'debug', verbose)
                    return df
                except KeyError:
                    log('ERROR: Data "{}" not found in store "{}"'.format(KEY, path_st), 'err', True)
                    return ts
        log('Store UPDATE: {:%c} , SIZE = {:.2f} KB. TOOK {:.3f} s'.format(ts, size_kb, time() - tic), 'debug', verbose)
        return ts
    except FileNotFoundError:
        log('ERROR: Store not found in {}'.format(path_st), 'err', True)
        return None


def delete_log_file(log_file, verbose=True):
    """
    Removes (logging) file from disk.

    :param log_file: :str: logging file path
    :param verbose: :bool: shows logging msgs in stdout.
    """
    if os.path.exists(log_file) and os.path.isfile(log_file):
        log('Deleting LOG FILE in {} ...'.format(log_file), 'warn', verbose, False)
        os.remove(log_file)
        return True
    return False


@timeit('extract_log_file')
def extract_log_file(log_file, extract_temps=True, verbose=True):
    """
    Extracts pd.DataFrame from logging file.

    :param log_file: :str: logging file path
    :param extract_temps: :bool: process RPI temperature logging entries (appends columns 'CPU' & 'GPU')
    :param verbose: :bool: shows logging msgs in stdout.
    :return: time-indexed pd.DataFrame
    """
    if os.path.exists(log_file):
        rg_log_msg = re.compile('(?P<tipo>INFO|WARNING|DEBUG|ERROR) \[(?P<func>.+?)\] '
                                '- (?P<ts>\d{1,2}/\d\d/\d\d\d\d \d\d:\d\d:\d\d): (?P<msg>.*?)\n', re.DOTALL)
        with open(log_file, 'r') as log_f:
            df_log = pd.DataFrame(rg_log_msg.findall(log_f.read()),
                                  columns=['tipo', 'func', 'ts', 'msg'])
        df_log.drop('func', axis=1, inplace=True)
        df_log['tipo'] = df_log['tipo'].astype('category')
        df_log['ts'] = df_log['ts'].apply(lambda x: dt.datetime.strptime(x, '%d/%m/%Y %H:%M:%S'))
        df_log.loc[df_log.msg.str.startswith('Tªs --> '), 'temp'] = True
        df_log.loc[df_log.msg.str.startswith('SENDED: '), 'debug_send'] = True
        b_warn = df_log['tipo'] == 'WARNING'
        df_log.loc[b_warn, 'no_red'] = df_log[b_warn].msg.str.startswith('OSError: [Errno 101]; C_UNREACHABLE:')
        # df_log.loc[b_warn, 'no_red'] = df_log[b_warn].msg.str.startswith('OSError:  La red es inaccesible')
        df_log['exec'] = df_log['msg'].str.contains(INIT_LOG_MARK).cumsum().astype(int)
        df_log = df_log.set_index('ts')
        if extract_temps:
            rg_temps = 'Tªs --> (?P<CPU>\d{1,2}\.\d) / (?P<GPU>\d{1,2}\.\d) ºC'
            df_log = df_log.join(df_log[df_log['temp'].notnull()].msg.str.extract(rg_temps, expand=True).astype(float))
        if verbose:
            clasific = df_log.groupby(['exec', 'tipo']).count().dropna(how='all').astype(int)
            log(clasific, 'ok', True, False)
            conteo_tipos = df_log.groupby('tipo').count()
            if 'ERROR' in conteo_tipos.index:
                log(df_log[df_log.tipo == 'ERROR'].dropna(how='all', axis=1), 'error', True, False)
            if 'INFO' in conteo_tipos.index:
                log(df_log[df_log.tipo == 'INFO'].dropna(how='all', axis=1), 'info', True, False)
        return df_log
    else:
        log("extract_log_file: '{}' doesn't exists".format(log_file), 'error', verbose, False)
        return pd.DataFrame([], columns=['tipo', 'func', 'ts', 'msg'])



# -*- coding: utf-8 -*-
"""
Flask forms

"""
from flask_wtf import FlaskForm
# from wtforms import StringField
# from wtforms.validators import DataRequired


class DummyForm(FlaskForm):
    """
    For post submits without any field...
    """
    pass


# class FileForm(FlaskForm):
#     """
#     File explorer
#     """
#     pathfile = StringField('pathfile', validators=[DataRequired()])
